<!doctype html>
<html lang="en">
<head>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">

    <title>mitmproxy</title>
    <link href="static/bootstrap.min.css" rel="stylesheet">
    <link href="static/mitmproxy.css" rel="stylesheet">
    <link rel="icon" href="static/images/favicon.ico" type="image/x-icon"/>
</head>

<body>
<nav class="navbar navbar-dark bg-dark mb-3">
    <div class="container">
        <div class="navbar-header">
            <div class="navbar-brand">
                <img src="static/images/mitmproxy-long.png" height="30" alt="mitmproxy logo" class="align-top"/>
            </div>
        </div>
    </div>
</nav>
<div class="container">
    {% block content %}
    {% endblock %}
</div>

</body>
</html>
