# Auto-generated by web/gen/web_columns.py
AVAILABLE_WEB_COLUMNS = [
    "icon",
    "index",
    "method",
    "version",
    "path",
    "quickactions",
    "size",
    "status",
    "time",
    "timestamp",
    "tls",
    "comment",
]
